@@ -135,7 +135,15 @@ end = struct
       try Some (Fs.find_in_path paths (name ^ ".ml")) with Not_found -> None)
 
   let read_event_list =
-    let read_paths ic : string list = input_value ic in
+    let rewrite_path path =
+      if Filename.is_relative path
+      then path
+      else
+        match Build_path_prefix_map.get_build_path_prefix_map () with
+        | Some map -> Build_path_prefix_map.rewrite (Build_path_prefix_map.flip map) path
+        | None -> path
+    in
+    let read_paths ic : string list = List.map (input_value ic) ~f:rewrite_path in
     fun { events_by_pc; units; pos_fname_to_source; toplevel = _; names; enabled }
         ~crcs
         ~includes
@@ -165,6 +173,11 @@ end = struct
                   try Some (Fs.find_in_path paths (Filename.basename pos_fname))
                   with Not_found -> find_ml_in_paths paths ev_module)
               in
+              let source =
+                match source with
+                | None -> None
+                | Some source -> Some (Fs.absolute_path source)
+              in
               if debug_sourcemap ()
               then
                 Format.eprintf
@@ -235,14 +248,7 @@ end = struct
             | _ -> loc.Location.loc_start
         in
         let src = unit.source in
-        Some
-          { Parse_info.name = Some pos.Lexing.pos_fname
-          ; src
-          ; line = pos.Lexing.pos_lnum - 1
-          ; col = pos.Lexing.pos_cnum - pos.Lexing.pos_bol
-          ; idx = 0
-          ; fol = None
-          }
+        Some (Parse_info.t_of_position ~src pos)
     with Not_found -> None
 
   let rec propagate l1 l2 =
@@ -371,7 +377,29 @@ module Constants : sig
 
   val inlined : Code.constant -> bool
 end = struct
-  let same_custom x y = Obj.field x 0 == Obj.field (Obj.repr y) 0
+  (* In order to check that two custom objects share the same kind, we
+     compare their identifier.  The identifier is currently extracted
+     from the marshaled value. *)
+  let ident_of_custom x =
+    (* Make sure tags are equal to custom_tag.
+       Note that in javascript [0l] and [0n] are not encoded as custom blocks. *)
+    if Obj.tag x <> Obj.custom_tag
+    then None
+    else
+      try
+        let bin = Marshal.to_string x [] in
+        match Char.code bin.[20] with
+        | 0x12 | 0x18 | 0x19 ->
+            let last = String.index_from bin 21 '\000' in
+            let name = String.sub bin ~pos:21 ~len:(last - 21) in
+            Some name
+        | _ -> assert false
+      with _ -> assert false
+
+  let same_ident x y =
+    match y with
+    | Some y -> String.equal x y
+    | None -> false
 
   let warn_overflow i i32 =
     warn
@@ -381,6 +409,12 @@ end = struct
       i32
       i32
 
+  let ident_32 = ident_of_custom (Obj.repr 0l)
+
+  let ident_64 = ident_of_custom (Obj.repr 0L)
+
+  let ident_native = ident_of_custom (Obj.repr 0n)
+
   let rec parse x =
     if Obj.is_block x
     then
@@ -391,17 +425,23 @@ end = struct
       then Float (Obj.magic x : float)
       else if tag = Obj.double_array_tag
       then Float_array (Obj.magic x : float array)
-      else if tag = Obj.custom_tag && same_custom x 0l
-      then Int (Obj.magic x : int32)
-      else if tag = Obj.custom_tag && same_custom x 0n
-      then (
-        let i : nativeint = Obj.magic x in
-        let i32 = Nativeint.to_int32 i in
-        let i' = Nativeint.of_int32 i32 in
-        if Poly.(i' <> i) then warn_overflow (Printf.sprintf "0x%nx (%nd)" i i) i32;
-        Int i32)
-      else if tag = Obj.custom_tag && same_custom x 0L
-      then Int64 (Obj.magic x : int64)
+      else if tag = Obj.custom_tag
+      then
+        match ident_of_custom x with
+        | Some name when same_ident name ident_32 -> Int (Obj.magic x : int32)
+        | Some name when same_ident name ident_native ->
+            let i : nativeint = Obj.magic x in
+            let i32 = Nativeint.to_int32 i in
+            let i' = Nativeint.of_int32 i32 in
+            if Poly.(i' <> i) then warn_overflow (Printf.sprintf "0x%nx (%nd)" i i) i32;
+            Int i32
+        | Some name when same_ident name ident_64 -> Int64 (Obj.magic x : int64)
+        | Some name ->
+            failwith
+              (Printf.sprintf
+                 "parse_bytecode: Don't know what to do with custom block (%s)"
+                 name)
+        | None -> assert false
       else if tag < Obj.no_scan_tag
       then
         Tuple (tag, Array.init (Obj.size x) ~f:(fun i -> parse (Obj.field x i)), Unknown)
@@ -422,6 +462,8 @@ end = struct
     | Int _ -> true
 end
 
+let const i = Constant (Int i)
+
 (* Globals *)
 type globals =
   { mutable vars : Var.t option array
@@ -628,13 +670,7 @@ module State = struct
   let pi_of_loc debug location =
     let pos = location.Location.loc_start in
     let src = Debug.find_source debug pos.Lexing.pos_fname in
-    { Parse_info.name = Some pos.Lexing.pos_fname
-    ; src
-    ; line = pos.Lexing.pos_lnum - 1
-    ; col = pos.Lexing.pos_cnum - pos.Lexing.pos_bol
-    ; idx = 0
-    ; fol = None
-    }
+    Parse_info.t_of_position ~src pos
 
   let rec name_rec debug i l s summary =
     match l, s with
@@ -725,6 +761,12 @@ let compiled_blocks = ref Addr.Map.empty
 
 let method_cache_id = ref 1
 
+let new_closure_repr =
+  (* true starting with 4.12 *)
+  match Ocaml_version.v with
+  | `V4_02 | `V4_03 | `V4_04 | `V4_06 | `V4_07 | `V4_08 | `V4_09 | `V4_10 | `V4_11 ->
+      false
+
 type compile_info =
   { blocks : Blocks.u
   ; code : string
@@ -763,7 +805,7 @@ let rec compile_block blocks debug_data code pc state =
     match last with
     | Branch (pc', _) | Poptrap ((pc', _), _) ->
         compile_block blocks debug_data code pc' state'
-    | Cond (_, _, (pc1, _), (pc2, _)) ->
+    | Cond (_, (pc1, _), (pc2, _)) ->
         compile_block blocks debug_data code pc1 state';
         compile_block blocks debug_data code pc2 state'
     | Switch (_, l1, l2) ->
@@ -784,7 +826,7 @@ and compile infos pc state instrs =
     else (
       State.name_vars state infos.debug pc;
       let stack = State.stack_vars state in
-      if debug_parser () then Format.eprintf "Branch %d (%a) @." pc print_var_list stack;
+      if debug_parser () then Format.eprintf "Branch %d (%a) @." pc Print.var_list stack;
       instrs, Branch (pc, stack), state)
   else (
     if debug_parser () then Format.eprintf "%4d " pc;
@@ -825,9 +867,9 @@ and compile infos pc state instrs =
         let x, state = State.fresh_var state in
         if debug_parser () then Format.printf "%a = 0@." Var.print x;
         (* We switch to a different block as this may have
-         changed the exception handler continuation *)
+           changed the exception handler continuation *)
         compile_block infos.blocks infos.debug code (pc + 2) state;
-        Let (x, Const 0l) :: instrs, Branch (pc + 2, State.stack_vars state), state
+        Let (x, const 0l) :: instrs, Branch (pc + 2, State.stack_vars state), state
     | ENVACC1 -> compile infos (pc + 1) (State.env_acc 1 state) instrs
     | ENVACC2 -> compile infos (pc + 1) (State.env_acc 2 state) instrs
     | ENVACC3 -> compile infos (pc + 1) (State.env_acc 3 state) instrs
@@ -847,7 +889,9 @@ and compile infos pc state instrs =
           infos
           (pc + 2)
           { state with
-            State.stack = State.Dummy :: State.Dummy :: State.Dummy :: state.State.stack
+            State.stack =
+              (* See interp.c *)
+              State.Dummy :: State.Dummy :: State.Dummy :: state.State.stack
           }
           instrs
     | APPLY ->
@@ -985,9 +1029,13 @@ and compile infos pc state instrs =
         let state = if nvars > 0 then State.push state else state in
         let vals, state = State.grab nvars state in
         let x, state = State.fresh_var state in
+        let env = List.map vals ~f:(fun x -> State.Var x) in
         let env =
-          Array.of_list (State.Dummy :: List.map vals ~f:(fun x -> State.Var x))
+          let code = State.Dummy in
+          let closure_info = State.Dummy in
+          if new_closure_repr then code :: closure_info :: env else code :: env
         in
+        let env = Array.of_list env in
         if debug_parser () then Format.printf "fun %a (" Var.print x;
         let nparams =
           match (get_instr_exn code addr).Instr.code with
@@ -1022,8 +1070,15 @@ and compile infos pc state instrs =
         done;
         let env = ref (List.map vals ~f:(fun x -> State.Var x)) in
         List.iter !vars ~f:(fun (i, x) ->
-            env := State.Var x :: !env;
-            if i > 0 then env := State.Dummy :: !env);
+            let code = State.Var x in
+            let closure_info = State.Dummy in
+            if new_closure_repr
+            then env := code :: closure_info :: !env
+            else env := code :: !env;
+            if i > 0
+            then
+              let infix_tag = State.Dummy in
+              env := infix_tag :: !env);
         let env = Array.of_list !env in
         let state = !state in
         let instrs =
@@ -1035,7 +1090,8 @@ and compile infos pc state instrs =
                 | GRAB -> getu code (addr + 1) + 1
                 | _ -> 1
               in
-              let state' = State.start_function state env (i * 2) in
+              let offset = if new_closure_repr then i * 3 else i * 2 in
+              let state' = State.start_function state env offset in
               let params, state' = State.make_stack nparams state' in
               if debug_parser () then Format.printf ") {@.";
               let state' = State.clear_accu state' in
@@ -1110,7 +1166,7 @@ and compile infos pc state instrs =
         let x, state = State.fresh_var state in
         if debug_parser () then Format.printf "%a = 0@." Var.print x;
         let instrs = register_global g i instrs in
-        compile infos (pc + 2) state (Let (x, Const 0l) :: instrs)
+        compile infos (pc + 2) state (Let (x, const 0l) :: instrs)
     | ATOM0 ->
         let x, state = State.fresh_var state in
         if debug_parser () then Format.printf "%a = ATOM(0)@." Var.print x;
@@ -1250,7 +1306,7 @@ and compile infos pc state instrs =
           infos
           (pc + 1)
           (State.pop 1 state)
-          (Let (x, Const 0l) :: Set_field (y, 0, z) :: instrs)
+          (Let (x, const 0l) :: Set_field (y, 0, z) :: instrs)
     | SETFIELD1 ->
         let y = State.accu state in
         let z = State.peek 0 state in
@@ -1261,7 +1317,7 @@ and compile infos pc state instrs =
           infos
           (pc + 1)
           (State.pop 1 state)
-          (Let (x, Const 0l) :: Set_field (y, 1, z) :: instrs)
+          (Let (x, const 0l) :: Set_field (y, 1, z) :: instrs)
     | SETFIELD2 ->
         let y = State.accu state in
         let z = State.peek 0 state in
@@ -1272,7 +1328,7 @@ and compile infos pc state instrs =
           infos
           (pc + 1)
           (State.pop 1 state)
-          (Let (x, Const 0l) :: Set_field (y, 2, z) :: instrs)
+          (Let (x, const 0l) :: Set_field (y, 2, z) :: instrs)
     | SETFIELD3 ->
         let y = State.accu state in
         let z = State.peek 0 state in
@@ -1283,7 +1339,7 @@ and compile infos pc state instrs =
           infos
           (pc + 1)
           (State.pop 1 state)
-          (Let (x, Const 0l) :: Set_field (y, 3, z) :: instrs)
+          (Let (x, const 0l) :: Set_field (y, 3, z) :: instrs)
     | SETFIELD ->
         let y = State.accu state in
         let z = State.peek 0 state in
@@ -1295,7 +1351,7 @@ and compile infos pc state instrs =
           infos
           (pc + 2)
           (State.pop 1 state)
-          (Let (x, Const 0l) :: Set_field (y, n, z) :: instrs)
+          (Let (x, const 0l) :: Set_field (y, n, z) :: instrs)
     | SETFLOATFIELD ->
         let y = State.accu state in
         let z = State.peek 0 state in
@@ -1307,7 +1363,7 @@ and compile infos pc state instrs =
           infos
           (pc + 2)
           (State.pop 1 state)
-          (Let (x, Const 0l) :: Set_field (y, n, z) :: instrs)
+          (Let (x, const 0l) :: Set_field (y, n, z) :: instrs)
     | VECTLENGTH ->
         let y = State.accu state in
         let x, state = State.fresh_var state in
@@ -1340,7 +1396,7 @@ and compile infos pc state instrs =
         in
         let x, state = State.fresh_var state in
         if debug_parser () then Format.printf "%a = 0@." Var.print x;
-        compile infos (pc + 1) (State.pop 2 state) (Let (x, Const 0l) :: instrs)
+        compile infos (pc + 1) (State.pop 2 state) (Let (x, const 0l) :: instrs)
     | GETSTRINGCHAR ->
         let y = State.accu state in
         let z = State.peek 0 state in
@@ -1383,7 +1439,7 @@ and compile infos pc state instrs =
         in
         let x, state = State.fresh_var state in
         if debug_parser () then Format.printf "%a = 0@." Var.print x;
-        compile infos (pc + 1) (State.pop 2 state) (Let (x, Const 0l) :: instrs)
+        compile infos (pc + 1) (State.pop 2 state) (Let (x, const 0l) :: instrs)
     | BRANCH ->
         let offset = gets code (pc + 1) in
         if debug_parser () then Format.printf "... (branch)@.";
@@ -1392,12 +1448,12 @@ and compile infos pc state instrs =
         let offset = gets code (pc + 1) in
         let x = State.accu state in
         let args = State.stack_vars state in
-        instrs, Cond (IsTrue, x, (pc + offset + 1, args), (pc + 2, args)), state
+        instrs, Cond (x, (pc + offset + 1, args), (pc + 2, args)), state
     | BRANCHIFNOT ->
         let offset = gets code (pc + 1) in
         let x = State.accu state in
         let args = State.stack_vars state in
-        instrs, Cond (IsTrue, x, (pc + 2, args), (pc + offset + 1, args)), state
+        instrs, Cond (x, (pc + 2, args), (pc + offset + 1, args)), state
     | SWITCH ->
         if debug_parser () then Format.printf "switch ...@.";
         let sz = getu code (pc + 1) in
@@ -1427,6 +1483,7 @@ and compile infos pc state instrs =
           (pc + 2)
           { (State.push_handler state x addr) with
             State.stack =
+              (* See interp.c *)
               State.Dummy
               :: State.Dummy
               :: State.Dummy
@@ -1585,12 +1642,12 @@ and compile infos pc state instrs =
           | _ -> assert false
         in
         if debug_parser () then Format.printf "%a = %ld@." Var.print x n;
-        compile infos (pc + 1) state (Let (x, Const n) :: instrs)
+        compile infos (pc + 1) state (Let (x, const n) :: instrs)
     | CONSTINT ->
         let n = gets32 code (pc + 1) in
         let x, state = State.fresh_var state in
         if debug_parser () then Format.printf "%a = %ld@." Var.print x n;
-        compile infos (pc + 2) state (Let (x, Const n) :: instrs)
+        compile infos (pc + 2) state (Let (x, const n) :: instrs)
     | (PUSHCONST0 | PUSHCONST1 | PUSHCONST2 | PUSHCONST3) as cc ->
         let state = State.push state in
         let x, state = State.fresh_var state in
@@ -1603,13 +1660,13 @@ and compile infos pc state instrs =
           | _ -> assert false
         in
         if debug_parser () then Format.printf "%a = %ld@." Var.print x n;
-        compile infos (pc + 1) state (Let (x, Const n) :: instrs)
+        compile infos (pc + 1) state (Let (x, const n) :: instrs)
     | PUSHCONSTINT ->
         let state = State.push state in
         let n = gets32 code (pc + 1) in
         let x, state = State.fresh_var state in
         if debug_parser () then Format.printf "%a = %ld@." Var.print x n;
-        compile infos (pc + 2) state (Let (x, Const n) :: instrs)
+        compile infos (pc + 2) state (Let (x, const n) :: instrs)
     | NEGINT ->
         let y = State.accu state in
         let x, state = State.fresh_var state in
@@ -1824,7 +1881,7 @@ and compile infos pc state instrs =
           infos
           (pc + 2)
           state
-          (Let (x, Prim (Extern "%int_add", [ Pv y; Pv z ])) :: Let (z, Const n) :: instrs)
+          (Let (x, Prim (Extern "%int_add", [ Pv y; Pv z ])) :: Let (z, const n) :: instrs)
     | OFFSETREF ->
         let n = gets code (pc + 1) in
         let x = State.accu state in
@@ -1832,7 +1889,7 @@ and compile infos pc state instrs =
         let instrs = Offset_ref (x, n) :: instrs in
         let x, state = State.fresh_var state in
         if debug_parser () then Format.printf "x = 0@.";
-        compile infos (pc + 2) state (Let (x, Const 0l) :: instrs)
+        compile infos (pc + 2) state (Let (x, const 0l) :: instrs)
     | ISINT ->
         let y = State.accu state in
         let x, state = State.fresh_var state in
@@ -1843,49 +1900,73 @@ and compile infos pc state instrs =
         let offset = gets code (pc + 2) in
         let x = State.accu state in
         let args = State.stack_vars state in
-        instrs, Cond (CEq n, x, (pc + offset + 2, args), (pc + 3, args)), state
+        let y = Var.fresh () in
+        ( Let (y, Prim (Eq, [ Pc (Int n); Pv x ])) :: instrs
+        , Cond (y, (pc + offset + 2, args), (pc + 3, args))
+        , state )
     | BNEQ ->
         let n = gets32 code (pc + 1) in
         let offset = gets code (pc + 2) in
         let x = State.accu state in
         let args = State.stack_vars state in
-        instrs, Cond (CEq n, x, (pc + 3, args), (pc + offset + 2, args)), state
+        let y = Var.fresh () in
+        ( Let (y, Prim (Eq, [ Pc (Int n); Pv x ])) :: instrs
+        , Cond (y, (pc + 3, args), (pc + offset + 2, args))
+        , state )
     | BLTINT ->
         let n = gets32 code (pc + 1) in
         let offset = gets code (pc + 2) in
         let x = State.accu state in
         let args = State.stack_vars state in
-        instrs, Cond (CLt n, x, (pc + offset + 2, args), (pc + 3, args)), state
+        let y = Var.fresh () in
+        ( Let (y, Prim (Lt, [ Pc (Int n); Pv x ])) :: instrs
+        , Cond (y, (pc + offset + 2, args), (pc + 3, args))
+        , state )
     | BLEINT ->
         let n = gets32 code (pc + 1) in
         let offset = gets code (pc + 2) in
         let x = State.accu state in
         let args = State.stack_vars state in
-        instrs, Cond (CLe n, x, (pc + offset + 2, args), (pc + 3, args)), state
+        let y = Var.fresh () in
+        ( Let (y, Prim (Le, [ Pc (Int n); Pv x ])) :: instrs
+        , Cond (y, (pc + offset + 2, args), (pc + 3, args))
+        , state )
     | BGTINT ->
         let n = gets32 code (pc + 1) in
         let offset = gets code (pc + 2) in
         let x = State.accu state in
         let args = State.stack_vars state in
-        instrs, Cond (CLe n, x, (pc + 3, args), (pc + offset + 2, args)), state
+        let y = Var.fresh () in
+        ( Let (y, Prim (Le, [ Pc (Int n); Pv x ])) :: instrs
+        , Cond (y, (pc + 3, args), (pc + offset + 2, args))
+        , state )
     | BGEINT ->
         let n = gets32 code (pc + 1) in
         let offset = gets code (pc + 2) in
         let x = State.accu state in
         let args = State.stack_vars state in
-        instrs, Cond (CLt n, x, (pc + 3, args), (pc + offset + 2, args)), state
+        let y = Var.fresh () in
+        ( Let (y, Prim (Lt, [ Pc (Int n); Pv x ])) :: instrs
+        , Cond (y, (pc + 3, args), (pc + offset + 2, args))
+        , state )
     | BULTINT ->
         let n = getu32 code (pc + 1) in
         let offset = gets code (pc + 2) in
         let x = State.accu state in
         let args = State.stack_vars state in
-        instrs, Cond (CUlt n, x, (pc + offset + 2, args), (pc + 3, args)), state
+        let y = Var.fresh () in
+        ( Let (y, Prim (Ult, [ Pc (Int n); Pv x ])) :: instrs
+        , Cond (y, (pc + offset + 2, args), (pc + 3, args))
+        , state )
     | BUGEINT ->
         let n = getu32 code (pc + 1) in
         let offset = gets code (pc + 2) in
         let x = State.accu state in
         let args = State.stack_vars state in
-        instrs, Cond (CUlt n, x, (pc + 3, args), (pc + offset + 2, args)), state
+        let y = Var.fresh () in
+        ( Let (y, Prim (Ult, [ Pc (Int n); Pv x ])) :: instrs
+        , Cond (y, (pc + 3, args), (pc + offset + 2, args))
+        , state )
     | ULTINT ->
         let y = State.accu state in
         let z = State.peek 0 state in
@@ -1944,7 +2025,7 @@ and compile infos pc state instrs =
              , Prim
                  ( Extern "caml_get_public_method"
                  , [ Pv obj; Pv tag; Pc (Int (Int32.of_int cache)) ] ) )
-          :: Let (tag, Const n)
+          :: Let (tag, const n)
           :: instrs)
     | GETDYNMET ->
         let tag = State.accu state in
@@ -2045,7 +2126,7 @@ let parse_bytecode code globals debug_data =
   tagged_blocks := Addr.Set.empty;
   let free_pc = String.length code / 4 in
   let blocks = match_exn_traps blocks in
-  0, blocks, free_pc
+  { start = 0; blocks; free_pc }
 
 (* HACK - override module *)
 
@@ -2063,46 +2144,6 @@ let override_global =
           :: Let (init_mod, jsmodule "CamlinternalMod" "init_mod")
           :: Let (update_mod, jsmodule "CamlinternalMod" "update_mod")
           :: instrs ) )
-  ; ( "Int_misc" (* Until https://github.com/ocaml/num/pull/16 is merged and released *)
-    , fun orig instrs ->
-        let x = Var.fresh_n "int_misc" in
-        let a0 = Var.fresh () in
-        let a1 = Var.fresh () in
-        let a2 = Var.fresh () in
-        let a3 = Var.fresh () in
-        let a_length_of_int = Var.fresh () in
-        let a_biggest_int = Var.fresh () in
-        let a_least_int = Var.fresh () in
-        let a_monster_int = Var.fresh () in
-        let length_of_int = 31 in
-        let monster_int = Int32.shift_left 1l length_of_int in
-        let biggest_int = Int32.sub monster_int 1l in
-        let least_int = Int32.neg biggest_int in
-        let map x i = Let (x, Field (orig, i)) in
-        ( x
-        , Let
-            ( x
-            , Block
-                ( 0
-                , [| a0
-                   ; a1
-                   ; a2
-                   ; a3
-                   ; a_length_of_int
-                   ; a_biggest_int
-                   ; a_least_int
-                   ; a_monster_int
-                  |]
-                , NotArray ) )
-          :: Let (a_monster_int, Const monster_int)
-          :: Let (a_least_int, Const least_int)
-          :: Let (a_biggest_int, Const biggest_int)
-          :: Let (a_length_of_int, Const (Int32.of_int length_of_int))
-          :: map a3 3
-          :: map a2 2
-          :: map a1 1
-          :: map a0 0
-          :: instrs ) )
   ]
 
 (* HACK END *)
@@ -2532,7 +2573,7 @@ let predefined_exceptions () =
         let v_name_js = Var.fresh () in
         let v_index = Var.fresh () in
         [ Let (v_name, Constant (String name))
-        ; Let (v_name_js, Prim (Extern "caml_js_from_string", [ Pc (IString name) ]))
+        ; Let (v_name_js, Constant (IString name))
         ; Let (v_index, Constant (Int (Int32.of_int (-index))))
         ; Let (exn, Block (248, [| v_name; v_index |], NotArray))
         ; Let
@@ -2544,4 +2585,4 @@ let predefined_exceptions () =
     |> List.concat
   in
   let block = { params = []; handler = None; body; branch = Stop } in
-  0, Addr.Map.singleton 0 block, 1
+  { start = 0; blocks = Addr.Map.singleton 0 block; free_pc = 1 }
