@@ -658,8 +658,8 @@ and element =
 
     method classList : tokenList t readonly_prop
 
-    (* Not supported by IE9 by default. Add +classList.js to the
-       Js_of_ocaml command line for compatibility *)
+    method closest : js_string t -> element t opt meth
+
     method style : cssStyleDeclaration t prop
 
     method innerHTML : js_string t prop
@@ -2333,9 +2333,9 @@ module Event : sig
 
   val scroll : event t typ
 
-  val focus : event t typ
+  val focus : focusEvent t typ
 
-  val blur : event t typ
+  val blur : focusEvent t typ
 
   val load : event t typ
 
